#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "fq_arith.h"
#include "monomial_mat.h"
#include "codes.h"
#include "LESS.h"
#include "rng.h"
#include "api.h"

#include "test_helpers.c"

#define GRN "\e[0;32m"
#define WHT "\e[0;37m"

#ifdef N_pad
#define NN N_pad
#else
#define NN N
#endif

/// taken from the nist package.
void fprintBstr(FILE *fp, char *S, unsigned char *A, unsigned long long L) {
    unsigned long long  i;
    fprintf(fp, "%s", S);
    for ( i=0; i<L; i++ )
        fprintf(fp, "%02X", A[i]);
    if ( L == 0 )
        fprintf(fp, "00");
    fprintf(fp, "\n");
}

/* Exhaustive testing of inverses mod Q */
void inverse_mod_tester(void){
    uint32_t value[Q-1];
    uint32_t inverse[Q-1];
    for(uint32_t i=1; i <= Q-1; i++){
        value[i-1] = i;
        inverse[i-1] = fq_inv(i);
    }
    int all_ok = 1;
    for(uint32_t i=1; i <= Q-1; i++){
        if((value[i-1]*inverse[i-1]) % Q !=1){
           printf("%u*%u=%u\n",
                  value[i-1],
                  inverse[i-1],
                  (value[i-1]*inverse[i-1])%Q);
           all_ok = 0;
        }
    }
    if (all_ok){
        puts("All inverses on F_q ok");
    }
}

/*
 *
 */
void rref_gen_compress_tester(void){
     generator_mat_t G = {0}, Gcheck;
     rref_generator_mat_t SF_G;
     uint8_t is_pivot_column[NN];

     /* randomly generate a non-singular G */
     do {
         generator_rnd(&G);
         memset(is_pivot_column,0,sizeof(is_pivot_column));
     } while ( generator_RREF(&G,is_pivot_column) == 0);

     memcpy(&Gcheck,&G, sizeof(G));
     generator_rref_compact(&SF_G,&G,is_pivot_column);
     generator_rnd(&G); /* fill with garbage to elicit faults */
     generator_rref_expand(&G,&SF_G);

    if( memcmp( &Gcheck,&G,sizeof(generator_mat_t)) !=0 ){
        printf("Generator SF compression: ko\n");
       fprintf(stderr," Comp-decomp\n");
       generator_pretty_print_name("G",&G);
       fprintf(stderr,"is_pivot = \n [ ");
       for(int x=0;x < N ;x++){fprintf(stderr," %d ",is_pivot_column[x]); }
       fprintf(stderr,"]\n");

       generator_rref_pretty_print_name("RREF-G",&SF_G);

       fprintf(stderr," Reference\n");
       generator_pretty_print_name("Gcheck",&Gcheck);
    } else {
        printf("Generator SF compression: ok\n");
    }
}

/*
 *
 */
void rref_gen_byte_compress_tester(void){
     generator_mat_t G = {0}, Gcheck;
     uint8_t G_compressed [RREF_MAT_PACKEDBYTES];
     uint8_t is_pivot_column[NN];

     /* randomly generate a non-singular G */
     do {
         generator_rnd(&G);
         memset(is_pivot_column,0,sizeof(is_pivot_column));
     } while ( generator_RREF(&G,is_pivot_column) == 0);

     memcpy(&Gcheck,&G, sizeof(G));
     compress_rref(G_compressed,&G,is_pivot_column);
     generator_rnd(&G); /* fill with garbage to elicit faults */
     expand_to_rref(&G,G_compressed,is_pivot_column);

    if( memcmp( &Gcheck,&G,sizeof(generator_mat_t)) !=0 ){
        printf("Generator SF byte compression: ko\n");
       fprintf(stderr," Comp-decomp\n");
       generator_pretty_print_name("G",&G);

       fprintf(stderr,"is_pivot = \n [ ");
       for(int x=0;x < N ;x++){fprintf(stderr," %d ",is_pivot_column[x]); }
       fprintf(stderr,"]\n");

       fprintf(stderr," \n\n\n\n\n\n\n\n\nReference\n");
       generator_pretty_print_name("Gcheck",&Gcheck);
    } else {
        printf("Generator SF compression: ok\n");
    }
}

/*
 *
 */
void info(void){
    fprintf(stderr,"Code parameters: n= %d, k= %d, q=%d\n", N,K,Q);
    fprintf(stderr,"num. keypairs = %d\n",NUM_KEYPAIRS);
    fprintf(stderr,"Fixed weight challenge vector: %d rounds, weight %d \n",T,W);
    fprintf(stderr,"Private key: %luB\n", sizeof(prikey_t));
    fprintf(stderr,"Public key %luB, %.3f kiB\n", sizeof(pubkey_t), ((float) sizeof(pubkey_t))/1024);
    fprintf(stderr,"Signature: %luB, %.3f kiB\n", sizeof(sign_t), ((float) sizeof(sign_t))/1024);

}


#define NUMBER_OF_TESTS 1
#define MLEN 160

/* returns 1 if the test is successful, 0 otherwise */
int LESS_sign_verify_test_multiple(void){
    unsigned long long smlen = 0, mlen1;
    unsigned char *m  = (unsigned char *)calloc(MLEN, sizeof(unsigned char));
    unsigned char *sm = (unsigned char *)calloc(MLEN+CRYPTO_BYTES, sizeof(unsigned char));
    unsigned char       pk[CRYPTO_PUBLICKEYBYTES] = {0}, sk[CRYPTO_SECRETKEYBYTES] = {0};

    int ret = 0;
    for (size_t i = 0; i < NUMBER_OF_TESTS; ++i) {
        const uint32_t msg_len = (uint32_t)rand() % MLEN;
        randombytes(m, msg_len);

        int ret_val;
        if ((ret_val = crypto_sign_keypair(pk, sk)) != 0) {
            printf("crypto_sign_keypair returned <%d>\n", ret_val);
            return -1;
        }

        fprintBstr(stdout, "pk = ", pk, CRYPTO_PUBLICKEYBYTES);
        fprintBstr(stdout, "sk = ", sk, CRYPTO_SECRETKEYBYTES);

        if ( (ret_val = crypto_sign(sm, &smlen, m, msg_len, sk)) != 0) {
            printf("crypto_sign returned <%d>\n", ret_val);
            return -1;
        }
        fprintBstr(stdout, "sm = ", sm, smlen);
        if ( (ret_val = crypto_sign_open(m, &mlen1, sm, smlen, pk)) != 0) {
            printf("crypto_sign_open returned <%d>\n", ret_val);
            return -1;
        }

        if(mlen1 != msg_len) {
            printf("crypto_sign_open wrong length\n");
            return -1;
        }
        printf("OK: %zu\n", i);

        ret |= ret_val;
    }

    return ret;
}


int LESS_sign_verify_test_KAT(void) {
    //unsigned char m[] = {0xD8,0x1C,0x4D,0x8D,0x73,0x4F,0xCB,0xFB,0xEA,0xDE,0x3D,0x3F,0x8A,0x03,0x9F,0xAA,0x2A,0x2C,0x99,0x57,0xE8,0x35,0xAD,0x55,0xB2,0x2E,0x75,0xBF,0x57,0xBB,0x55,0x6A,0xC8};
	//uint8_t seed[] = {0x06,0x15,0x50,0x23,0x4D,0x15,0x8C,0x5E,0xC9,0x55,0x95,0xFE,0x04,0xEF,0x7A,0x25,0x76,0x7F,0x2E,0x24,0xCC,0x2B,0xC4,0x79,0xD0,0x9D,0x86,0xDC,0x9A,0xBC,0xFD,0xE7,0x05,0x6A,0x8C,0x26,0x6F,0x9E,0xF9,0x7E,0xD0,0x85,0x41,0xDB,0xD2,0xE1,0xFF,0xA1};
    // uint8_t seed[] = {0xF1,0x90,0x2A,0x78,0x15,0xF3,0x7B,0xC7,0xF5,0x80,0x2D,0x8C,0xBC,0xE5,0xB4,0x8D,0x82,0xEB,0x85,0x69,0x17,0x18,0x06,0x2B,0xFB,0x84,0xD8,0xC0,0x6A,0xA4,0x1D,0x6E,0x90,0x39,0xB0,0xA1,0x07,0x24,0x5D,0xAF,0xA4,0xEC,0x10,0x9A,0x57,0x33,0x29,0x14};
    // uint8_t m[] = {0x1C,0xDF,0x0A,0xE1,0x12,0x47,0x80,0xA8,0xFF,0x00,0x31,0x8F,0x77,0x9A,0x3B,0x86,0xB3,0x50,0x4D,0x05,0x9C,0xA7,0xAB,0x3F,0xE4,0xD6,0xEA,0xE9,0xFD,0x46,0x42,0x8D,0x1D,0xAB,0xB7,0x04,0xC0,0x73,0x5A,0x8F,0xE8,0x70,0x8F,0x40,0x97,0x41,0x01,0x7B,0x72,0x3D,0x9A,0x30,0x4E,0x54,0xFD,0xC5,0x78,0x9A,0x7B,0x07,0x48,0xC2,0x46,0x4B,0x73,0x08,0xAC,0x96,0x65,0x11,0x56,0x44,0xC5,0x69,0xAE,0x25,0x3D,0x52,0x05,0x75,0x13,0x42,0x57,0x4C,0x03,0x34,0x6D,0xDD,0xC1,0x95,0x0A,0x62,0x73,0x54,0x66,0x16,0xB9,0x6D,0x0C,0x5E,0xCE,0x0A,0x04,0x4A,0xF0,0xED,0xEF,0xBE,0x44,0x5F,0x9A,0xE3,0x7D,0xA5,0xAF,0xB8,0xD2,0x2A,0x56,0xD9,0xFD,0x18,0x01,0x42,0x5A,0x0A,0x27,0x6F,0x48,0x43,0x1D,0x7A,0xF0,0x39,0x52,0x1E,0x54,0x95,0x51,0x48,0x13,0x91,0xFE,0x5F,0x4E,0xBF,0xB7,0x64,0x4D,0x9F,0x97,0x82,0xD8,0x3A,0x95,0x13,0x7E,0x84,0xEA,0x3A,0xEB,0x3C,0x2F,0x80,0x99};
    uint8_t seed[] ={0x83,0xC6,0x53,0x70,0x8F,0xAF,0x3E,0x5F,0x6F,0xBC,0x9D,0xFB,0xE6,0xFB,0x5E,0x83,0xE5,0x72,0xA7,0x68,0x86,0x45,0xD7,0x5D,0x2C,0x48,0x35,0xB2,0x86,0x95,0xDE,0xA4,0xBD,0x70,0x93,0x74,0x0D,0x0F,0xF4,0x32,0x37,0x35,0x4E,0xAD,0x1C,0x97,0x8B,0xC2};
    uint8_t m[] ={0xE3,0xB5,0x7B,0x20,0x83,0x52,0xA8,0x20,0xF6,0x22,0xA6,0x94,0xB7,0xC3,0xF6,0xF2,0x97,0x23,0x9E,0xF0,0xA0,0x69,0x61,0x5D,0xC6,0x64,0xC0,0x2F,0x18,0x22,0xBB,0xA4,0x8E,0x11,0xE3,0x7B,0xD9,0x74,0x9C,0x98,0xFA,0xCE,0xFF,0xFB,0x0F,0xE1,0x79,0x2A,0x38,0x6B,0xE1,0x0C,0xA7,0xB9,0x8C,0xC8,0x74,0xC6,0x8C,0x36,0xF5,0x09,0x6D,0x37,0x18,0xDC,0x93,0xE0,0x73,0x4D,0x6D,0x6F,0x91,0x3E,0x3B,0x95,0x8D,0xC1,0xFD,0x14,0x24,0x81,0x8C,0x94,0x37,0xB0,0xFD,0x59,0x72,0x8E,0xD4,0x6A,0x79,0xFB,0x52,0xC7,0x37,0xA1,0xD1,0xD2,0x6F,0x04,0xEB,0xAC,0x27,0x9A,0x7F,0xF6,0xA9,0x71,0xE2,0xB6,0x95,0x76,0xB7,0x12,0xD9,0x22,0x4E,0xA1,0x8F,0xB9,0xBF,0x4E,0x61,0x3A,0x89,0x35,0xF3,0xB3,0x6A,0x07,0x3B,0x01,0xF3,0x7B,0xDC,0x0B,0x77,0x98,0x1C,0x8F,0x28,0x04,0xE9,0x3C,0x39,0x54,0x19,0x35,0x2B,0x85,0xC8,0xA3,0x2D,0xD7,0x7D,0x41,0xDA,0x9B,0xF3,0xEC,0xB9,0x14,0x17,0x3E,0x80,0xDD,0x1F,0xC0,0x6E,0x8F,0xF5,0xBF,0x0E,0x4F,0x74,0x24,0x84,0x9A,0x15,0xEB,0x7F,0xAF,0x7D,0xE7,0x74,0x56,0xEB,0xB6,0x4D,0x10,0xDC,0x10,0xFE,0xC6,0x25,0x40,0x70,0xC7,0xDF,0x38,0x73,0x97,0x13,0x73,0x72,0xEA,0x3A,0x53,0xDF,0xDA,0x7D,0xA1,0x34,0x14,0xAF,0x2D,0xF1,0x6C,0x1E,0x38,0xC5,0xC7,0x0A,0x5F,0x5F,0x44,0xF7,0x25,0xD6,0x22,0x04,0x92,0x56,0xBB,0x15,0xDC,0x04,0xA8,0xD8,0x46,0xA1,0xA0,0xDA,0xE7,0xE7,0x65,0xA7,0xF0,0x0C,0x49,0x8F,0x1D,0x0B,0x28,0x93,0xB8,0x40,0x5B,0xE4,0xA4,0x3F,0xB7,0xE9,0x78,0x81,0x06,0x9A,0x49,0x13,0x4A,0x2A,0x84,0x71,0x84,0xB8,0x2E,0xB5,0xA6,0x90,0xD8,0x7B,0xAF,0x2F,0x57,0x96,0x19,0xEE,0x19,0xA3,0xD7,0xA7,0xC7,0xEE,0xA7,0x2D,0x6E,0x3F,0xCC,0xF0,0xA8,0x09,0x2B,0xB8,0xD3,0xC6,0xB5,0x51,0xF2,0x7E,0x63,0xE7,0x62,0xA3,0x0B,0x4A,0x4D,0xF2,0xDB,0xC4,0xD1,0x19,0x13,0x9A,0xE1,0xB1,0x35,0xD0,0x6F,0xF8,0x27,0x84,0x69,0x01,0x57,0x77,0x00,0x93,0x5E,0x00,0x11,0xB6,0x54,0x61,0xC2,0xEF,0x9A,0x7B,0x71,0xEE,0xA3,0x3C,0x8C,0xA4,0x51,0x9C,0x7B,0xCF,0xB5,0x57,0xC5,0xE1,0xD4,0x2D,0x92,0x43,0xF2,0xDC,0x34,0x05,0x7F,0x5E,0x0C,0xCB,0x9A,0x45,0x7F,0xC3,0x4D,0xCB,0x10,0xD9,0xB4,0x7F,0x6E,0xC3,0xB9,0x55,0x0D,0x3A,0xE4,0xFD,0x59,0x3D,0xFA,0x3E,0x28,0xC6,0xCC,0xA1,0xFF,0x1E,0xBC,0x9D,0x98,0xDA,0x8D,0xB8,0x69,0xF8,0xC8,0x0B,0xDB,0xF8,0xAD,0x46,0x84,0xAC,0xB6,0xA7,0x79,0xCA,0x9D,0x0A,0x10,0x6F,0x26,0xDA,0x17,0x04,0x37,0x73,0x86,0x26,0x81,0xC5,0xDD,0x2D,0xEB,0x1B,0xCA,0x2C,0xA4,0x8D,0x4F,0xBB,0x4B,0xB7,0xC1,0xF7,0x65,0xDC,0xA3,0xA1,0xD9,0x91,0xD8,0x90,0xB9,0xA8,0x75,0x1C,0xEA,0xFF,0x54,0x39,0x97,0xFA,0xE5,0xB1,0x28,0xAB,0x2E,0xF2,0x2B,0x3B,0xE9,0x44,0x99,0xDF,0xD9,0xD8,0xE7,0x8F,0xB4,0xC8,0x2C,0xA8,0xD2,0x96,0xB0,0x41,0x5E,0x84,0xCA,0x8B,0x5F,0x20,0x24,0x45,0x5B,0x5D,0xEC,0xC8,0xB4,0xCC,0xDC,0x7B,0xC4,0xEE,0x06,0xB4,0xF0,0xC6,0x6E,0x67,0x48,0xFB,0xD0,0x7E,0x3A,0x3B,0xC5,0xB4,0xB6,0x88,0x9C,0x40,0xDC,0x4A,0x97,0xAE,0x3E,0xB4,0x3C,0x39,0x14,0xDE,0xF9,0x76,0xEF,0xE3,0xBF,0xD8,0x4A,0x09,0x3B,0xD6,0x91,0x02,0xD7,0xB3,0x7C,0x89,0xB4,0x58,0xA5,0x5B,0x98,0xA1,0x97,0x4A,0x13,0xA7,0x68,0x5D,0x26,0xE9,0xD8,0x16,0xC7,0x95,0x85,0xBC,0xFC,0x10,0x42,0xC2,0xAF,0x88,0x53,0x4A,0x9F,0xE8,0xB0,0xA6,0xC8,0xC4,0x43,0x55,0xA6,0xD6,0x06,0xF9,0x02,0xDB,0x40,0xD5,0x49,0x02,0x64,0xBF,0x0F,0x35,0x2C,0x27,0x35,0x56,0x33,0xCB,0x09,0x52,0x68,0xD5,0xB8,0xBE,0xC9,0x85,0xA6,0x2D,0x84,0xB2,0x32,0x3F,0xE8,0x14,0x05,0x3F,0x05,0xDE,0xDC,0x22,0x02,0x9D,0x29,0x98,0xBD,0x0B,0xCB,0x25,0x5C,0x16,0x2C,0x4B,0xC0,0x3F,0x60,0xE3,0x58,0x0A,0xC3,0xAE,0x86,0xC3,0x78,0x50,0x11,0x0E,0x9A,0x1B,0xCB,0xD7,0x5F,0x64,0xA0,0xDD,0x60,0xB9,0x41,0xE2,0xF5,0x7D,0xA9,0xD7,0x24,0x98,0xB3,0xEA,0x83,0x24,0xEE,0xA5,0x3D,0xA3,0x89,0x55,0x85,0xED,0x29,0x42,0xB9,0x14,0x0F,0x26,0x08,0x95,0xDC,0x6A,0x11,0x31,0xA4,0xC3,0xAD,0x2B,0x64,0x02,0x8B,0xB8,0xC0,0xFD,0x67,0xE1,0xBE,0x4C,0x07,0xF8,0x08,0xB4,0x7D,0xAE,0xF3,0x06,0xFD,0x95,0x78,0x02,0x5F,0x9C,0x63,0x96,0x60,0x07,0x58,0x37,0xB2,0xC9,0x54,0x73,0xF7,0xF8,0x60,0xD6,0xEA,0x2C,0x53,0xF4,0xBA,0x67,0x7A,0x23,0x45,0xCF,0x21,0x2C,0x77,0x57,0xBB,0x94,0xF1,0xA4,0xF7,0x6D,0x4E,0x96,0x62,0x5F,0x6F,0xE0,0x51,0xB8,0x24,0x6D,0x1B,0x76,0x11,0xBF,0x6F,0xE3,0x25,0xFF,0xFF,0x85,0x14,0xD2,0xF9,0xA3,0x45,0x3F,0x0E,0x77,0xAE,0x8B,0x95,0x8A,0xB5,0xB5,0x67,0xE5,0x41,0xF1,0x56,0xC6,0xF4,0xD3,0x15,0xB4,0xC3,0xC5,0x47,0xD5,0x9B,0xBD,0x0D,0x74,0x03,0xE2,0xE6,0xA4,0x9B,0x9E,0x7D,0x3F,0xDB,0xA3,0x38,0xAD,0xA4,0x18,0x75,0xCE,0xB0,0x38,0x30,0xA8,0x46,0xA1,0xFB,0x26,0x6C,0x0F,0x12,0x28,0xAA,0xD2,0xB7,0x6A,0x2E,0x34,0x04,0x27,0x8D,0xBE,0x48,0x29,0x07,0x20,0x6F,0xA6,0x64,0x87,0xAD,0x2C,0x99,0x98,0x67,0xF8,0x70,0xC8,0xCB,0x7A,0x70,0xB8,0x34,0x37,0xE1,0x4B,0x9E,0x89,0x3B,0xF6,0xB3,0x91,0xDA,0xD7,0x5E,0x84,0x58,0x8E,0x88,0x22,0x46,0xD1,0x61,0x79,0x9A,0xDE,0xA6,0x3A,0xDF,0x1A,0xD7,0x06,0xC0,0xA3,0xB7,0x6B,0xAE,0x59,0x5D,0x84,0xB2,0x1A,0xE9,0xDA,0x30,0xBB,0xC0,0x85,0x69,0x87,0xF2,0xC2,0xC5,0x43,0xD9,0x77,0x74,0x7B,0x8C,0xBD,0x5A,0x61,0x3B,0x92,0x80,0x4E,0xCC,0x52,0x84,0xED,0x23,0x65,0x0E,0x9D,0xAF,0xB4,0xB7,0x6D,0x63,0xF0,0x69,0x71,0x08,0x97,0x33,0x4F,0x18,0xEA,0x6B,0x0C,0xBF,0x99,0xCD,0x59,0x0A,0x78,0xE3,0xB0,0x50,0xE1,0xBB,0x24,0xC8,0x6D,0x63,0x23,0xA1,0x71,0x06,0xF0,0xCA,0xE3,0xF3,0x0B,0x01,0xE4,0xEB,0x3D,0xB1,0xB5,0xF3,0xA4,0x77,0x1A,0x88,0x0C,0x8A,0xC0,0x6B,0xCD,0x5A,0x82,0xD4,0x10,0x3D,0x04,0x52,0xFD,0x7B,0x54,0x83,0x4C,0x1C,0xF8,0x59,0x5D,0xD7,0x7F,0x82,0xD4,0xAD,0x9E,0xBC,0x1C,0xFD,0x0C,0x9A,0x8C,0xC7,0x87,0xE1,0x0A,0xA4,0xD1,0x68,0x84,0x74,0x20,0x8B,0x69,0xFF,0x7A,0xD4,0xDA,0x69,0x86,0xE5,0xF6,0x2A,0x34,0xAC,0x30,0x93,0xE0,0xFB,0x1E,0xFE,0x8A,0xE3,0xA9,0x6F,0x6A,0xAE,0x09,0xB0,0xE8,0xF6,0xE7,0xA2,0xB6,0x5C,0x73,0x87,0x99,0x9C,0xEC,0xCA,0x43,0xCC,0x33,0xF0,0x26,0xDC,0x19,0xBB,0xFD,0x86,0x7C,0x48,0x12,0x7C,0xFF,0x57,0x9D,0x1D,0x71,0xAF,0xF0,0xC4,0xA0,0xE2,0x0F,0x9F,0xDF,0xD5,0x99,0xA6,0x16,0x9D,0xF1,0xB8,0x5F,0x60,0x51,0xE0,0x22,0x90,0xDF,0x6F,0x5E,0xDE,0x4F,0x29,0xBB,0x6F,0x0C,0x8F,0x80,0x6D,0x68,0x50,0xC6,0x53,0x4E,0xCD,0xDC,0xCD,0x75,0xBB,0x8E,0x4A,0x09,0x7C,0x70,0x44,0x55,0x85,0x74,0x0F,0x82,0x2E,0x5C,0xEB,0xB0,0xE1,0x9E,0xAC,0x82,0xBB,0x78,0xEB,0xDE,0x2C,0xA6,0x0A,0x81,0x0A,0xC6,0xC5,0x41,0x19,0xFD,0x64,0x27,0xDA,0x8A,0x01,0x55,0xEF,0x48,0x65,0x35,0x15,0xA9,0x19,0xB2,0x99,0xA3,0x06,0xFD,0x3C,0x62,0xB5,0x05,0xA6,0x91,0x1D,0xB2,0xB5,0x6C,0xA2,0xF2,0x96,0xE4,0x87,0xBA,0x02,0xC5,0x46,0xEC,0xA2,0x78,0x3A,0xDE,0x8E,0x46,0xA8,0xC7,0x8E,0xB1,0xF3,0xD7,0xC0,0x4B,0xB2,0x45,0x48,0xF9,0x23,0x83,0xE4,0x75,0xCE,0x6E,0x57,0x2D,0x8D,0xE1,0xBF,0xA9,0xB3,0xE3,0x5D,0x9B,0xD6,0xC7,0x95,0x47,0xB5,0x92,0xC9,0x56,0x93,0x75,0x00,0x10,0xA3,0xD2,0x2C,0xBB,0x31,0xAA,0x5A,0x4A,0xBE,0x94,0x89,0x78,0x31,0xB1,0xED,0x92,0x87,0x63,0x1F,0x00,0x6A,0x73,0x5C,0x36,0xBC,0x84,0xA8,0xC8,0x74,0x97,0xEE,0xA4,0x87,0x38,0x01,0xA7,0x33,0xF3,0x5B,0x32,0x8C,0x7D,0x2C,0xCB,0xE4,0xA4,0x1C,0x19,0x3D,0x22,0xF9,0x72,0x57,0x1B,0xA7,0x63,0x0B,0x33,0x08,0x07,0x93,0x49,0x8C,0xC8,0x5E,0x6E,0xEA,0x1C,0x41,0x29,0x14,0x45,0x9D,0xA1,0x75,0xA6,0xDB,0x86,0x58,0xD0,0xBD,0x7A,0x82,0x3F,0xAB,0x28,0x6E,0xDC,0x20,0xC7,0x85,0xC4,0x0B,0xFD,0x53,0x99,0x24,0xA2,0x4A,0xF4,0xE3,0xD3,0x7B,0xD7,0x81,0x35,0x36,0x77,0xC7,0x6D,0x46,0x72,0x09,0x8F,0x5B,0xDD,0x17,0x01,0x70,0x12,0x57,0x1D,0x9A,0xFD,0xA0,0x5A,0x40,0xAB,0x56,0x99,0x8E,0x40,0xF5,0xE3,0x59,0xC4,0x3D,0xFE,0x32,0xCA,0x10,0xA4,0x5B,0xF0,0x8F,0x67,0xD1,0x28,0xC2,0x4B,0x1A,0xCC,0x03,0xCB,0xAC,0x46,0xBA,0x6C,0xA5,0xA5,0x32,0xC1,0x05,0xE9,0x1E,0x0C,0x77,0xED,0x59,0xFB,0x53,0x4A,0xEE,0xCD,0x68,0x73,0x5A,0x49,0x78,0x17,0x7B,0xB5,0xA6,0x56,0xB9,0xF8,0x3B,0x20,0x2B,0xB6,0x04,0xD6,0x1A,0x24,0x57,0x4C,0x16,0x65,0x6E,0x51,0x2C,0x0A,0x4C,0xC6,0xF5,0x97,0xB3,0x26,0x85,0x73,0xE1,0x05,0x39,0xD1,0xBA,0x77,0x5E,0xD8,0x3B,0xB6,0x80,0xBB,0x91,0x15,0x01,0x1C,0x6A,0xD4,0x3F,0xBB,0x66,0xFB,0x37,0xC4,0x67,0x24,0x90,0x60,0xA1,0x58,0x6D,0xF2,0x7B,0x2C,0xEF,0xA6,0x52,0x65,0xCC,0xB9,0x05,0x1E,0x46,0x80,0x00,0xCC,0xAE,0x24,0xF0,0x8B,0xA9,0x41,0xA8,0x18,0x0A,0x64,0xBB,0x62,0x4F,0x14,0x6C,0x8E,0xC5,0x62,0x36,0x3B,0x32,0xC3,0x69,0xF6,0x29,0x97,0xC4,0xB1,0x37,0x5D,0xD7,0xDE,0x64,0x72,0x5A,0x59,0x85,0x29,0x24,0x42,0x73,0xCA,0xF8,0x39,0x89,0x13,0xC6,0xFC,0x01,0x52,0x26,0x83,0xCF,0x1F,0x9F,0x96,0x5C,0x49,0x1A,0xBE,0x7A,0x55,0x4F,0x00,0x19,0x51,0x4E,0xD9,0x8D,0x75,0xEB,0x8B,0xB8,0x56,0x5F,0x77,0xC1,0x95,0xF6,0x29,0xF9,0x81,0x63,0x49,0x4B,0x4A,0xA2,0x67,0x4F,0x92,0xA4,0x1D,0xCB,0x67,0xED,0xD1,0xD8,0x18,0xA5,0xB9,0x89,0x93,0xD0,0xB1,0x19,0x8B,0xB6,0xBE,0xDA,0xBB,0xB4,0x86,0xBC,0x6F,0xDE,0x03,0x94,0x33,0xE8,0x42,0xBA,0xC5,0x68,0xA5,0xB4,0xEA,0xCC,0x02,0x8C,0xC2,0x54,0x4B,0x57,0xD8,0x88,0x38,0x48,0xDD,0xDE,0xE2,0xE9,0x67,0xEA,0x85,0xA6,0x10,0x2B,0xD0,0xAB,0xDD,0xA4,0x1C,0x3D,0x78,0x44,0x7B,0xEE,0x1D,0x49,0x49,0x44,0x9A,0xBA,0xA9,0xB3,0x37,0x7E,0x8C,0xED,0xCF,0x04,0xA5,0x00,0xFD,0x1A,0x69,0x16,0xE2,0x69,0x83,0xE6,0x4B,0x5E,0x96,0xFE,0xF8,0x7B,0x32,0xA0,0x60,0x44,0x4D,0x37,0x44,0x09,0x26,0x24,0x53,0xCB,0x13,0x76,0xC3,0x49,0xA8,0xB5,0xD1,0x76,0x7B,0x1E,0x29,0x91,0xA1,0xA6,0x04,0x4E,0x0F,0x58,0x83,0x1B,0xD1,0x1F,0x12,0x15,0x96,0x75,0xD2,0x15,0xD7,0xEA,0xA7,0x48,0x07,0xC9,0x95,0xFE,0x22,0x01,0x7E,0x30,0x48,0x2D,0xB8,0xA4,0xB0,0x9C,0xA7,0x80,0x08,0x22,0xC7,0x5C,0x92,0xFF,0x64,0x9F,0xC0,0x72,0x8F,0x5A,0x1D,0x44,0xEF,0xE7,0xD0,0xFF,0x14,0x72,0x74,0x15,0x2D,0x5F,0x2F,0x60,0x34,0x2C,0x8F,0x5F,0x95,0x1D,0x8C,0x95,0xF8,0x3C,0x1D,0x54,0x61,0x3A,0x18,0x2D,0x9D,0xCA,0x68,0xF5,0x4F,0xD5,0x50,0x47,0xF1,0xF9,0x0C,0xFE,0xCC,0x04,0xD7,0x33,0xDF,0xA8,0x2C,0xFF,0x26,0x18,0xF2,0x9A,0x4D,0xB4,0xF7,0xE1,0xE5,0x9D,0xEA,0xD5,0x8C,0xA6,0x5D,0x07,0xCC,0x90,0xC2,0x5F,0x80,0x4A,0x89,0x5D,0x6A,0x82,0xF9,0x37,0x54,0x51,0xCC,0x55,0x50,0x6D,0x27,0x6F,0xBF,0x78,0x3F,0x7D,0x4D,0x53,0xB9,0xBF,0xB8,0x3D,0xBE,0x4A,0x87,0x71,0xAF,0xE2,0x1A,0xC5,0x43,0x98,0x3D,0x68,0x03,0x4B,0xAD,0xC9,0x80,0xF9,0x43,0x45,0x27,0xF9,0xED,0xAA,0x2E,0x22,0x86,0x46,0xFD,0xF7,0x5B,0x44,0x89,0x9E,0x74,0x9C,0xF4,0xC9,0xE5,0xB3,0x45,0x22,0x23,0x85,0xA4,0x42,0x43,0x82,0x60,0x3A,0xD6,0xEF,0xC2,0x4C,0x56,0xE7,0x69,0x02,0x8F,0x43,0x94,0xF2,0xF6,0x22,0x0A,0x9B,0x39,0x0D,0x39,0x5E,0x41,0x24,0x98,0xE5,0x7A,0x08,0xBA,0xD9,0x27,0xB8,0xBD,0x5D,0x76,0xE1,0x8E,0x8F,0xEB,0x45,0x7F,0xCB,0xD3,0x24,0x8D,0x21,0x82,0x36,0xB0,0x77,0x83,0xE5,0x7F,0xBF,0xA0,0x3C,0x29,0x2A,0x9F,0x57,0x19,0xE6,0xAE,0xF2,0xEE,0xA3,0xFA,0xB2,0xCA,0xEE,0xD5,0x44,0x2E,0x89,0xBF,0xFB,0x23,0x6C,0xB1,0x3D,0xB2,0xCF,0x9C,0x35,0xA3,0x8C,0x33,0x8C,0x37,0x7C,0x47,0x5D,0xAF,0x45,0xF8,0xEA,0x82,0x2F,0x9A,0xAA,0xC1,0x34,0x25,0xFB,0xD4,0x3D,0x3D,0xD9,0x22,0x93,0x67,0xF0,0xB3,0x68,0x7D,0x7E,0x82,0xAC,0x5E,0xC2,0xFC,0x7C,0xDB,0x69,0xC9,0x9A,0x4E,0xB1,0xB8,0xE4,0x54,0x65,0xC6,0xA5,0x3F,0x16,0xAC,0x0C,0x4E,0x0C,0x97,0x0B,0x8C,0x73,0x2A,0xF5,0x15,0xC0,0x9E,0xAF,0x25,0x59,0x6F,0x64,0xA0,0x4A,0xE4,0x62,0x10,0x37,0xB8,0x84,0x1F,0xD2,0xB1,0xBB,0xCB,0x31,0x0E,0xA2,0x3E,0x12,0x2B,0x0B,0x9A,0xB9,0x6D,0x8F,0x77,0x02,0x95,0x2D,0x0E,0x96,0xE4,0xCF,0x79,0xC2,0xA3,0x0D,0xF0,0x09,0x1A,0xCD,0xA9,0x14,0x79,0xEE,0x29,0x79,0xB0,0x05,0x49,0x97,0xC4,0x8F,0x6A,0x0E,0x90,0x9B,0xC5,0x2A,0x94,0x34,0x59,0xAF,0x25,0x55,0x39,0x69,0xEB,0x31,0xCE,0x76,0x85,0x36,0x9A,0x7F,0xB0,0x14,0x56,0x1B,0x46,0x97,0xB8,0xBC,0xE2,0x20,0x98,0x31,0x36,0xE5,0xEB,0x23,0x03,0xCC,0xA4,0xEA,0xDD,0x4C,0x6C,0xC7,0x4E,0xA2,0xFE,0x69,0xD4,0x48,0xAE,0x6E,0xD9,0x53,0xA8,0x03,0x63,0xDD,0xED,0x55,0x91,0xB2,0x7A,0x1E,0xA9,0x56,0xDF,0x08,0x1C,0xE9,0x9A,0xA5,0x9D,0xFC,0x78,0x9D,0x9D,0x8F,0xAE,0x95,0x2B,0x07,0x37,0x09,0x9D,0x46,0x7D};

    //for (uint32_t i=0; i<48; i++) {
    //    seed[i] = i;
	//}

    // init_randombytes(seed, 48);
    initialize_csprng(&platform_csprng_state,
                      (const unsigned char *)seed,
                      48);

    const uint32_t mlen = sizeof(m);
    unsigned long long smlen = 0, mlen1;
    unsigned char *m1 = (unsigned char *)calloc(mlen+CRYPTO_BYTES, sizeof(unsigned char));
    unsigned char *sm = (unsigned char *)calloc(mlen+CRYPTO_BYTES, sizeof(unsigned char));
    unsigned char pk[CRYPTO_PUBLICKEYBYTES] = {0}, sk[CRYPTO_SECRETKEYBYTES] = {0};

    int ret_val;
    if ((ret_val = crypto_sign_keypair(pk, sk)) != 0) {
        printf("crypto_sign_keypair returned <%d>\n", ret_val);
        return -1;
    }
    fprintBstr(stdout, "pk = ", pk, CRYPTO_PUBLICKEYBYTES);
    fprintBstr(stdout, "sk = ", sk, CRYPTO_SECRETKEYBYTES);

    if ( (ret_val = crypto_sign(sm, &smlen, m, mlen, sk)) != 0) {
        printf("crypto_sign returned <%d>\n", ret_val);
        return -1;
    }

    fprintBstr(stdout, "sm = ", sm, smlen);
    if ( (ret_val = crypto_sign_open(m1, &mlen1, sm, smlen, pk)) != 0) {
        printf("crypto_sign_open returned <%d>\n", ret_val);
        return -1;
    }

    free(m1);
    free(sm);
    printf("all good\n");
    return 0;
}


#define NUM_TEST_ITERATIONS 10
int main(int argc, char* argv[]){
    (void)argc;
    (void)argv;
    // LESS_sign_verify_test_multiple();
    return LESS_sign_verify_test_KAT();
}
